# https://stackoverflow.com/a/72126394/

"""
multithreading python example
count the number of primes up until a max integer
"""

import multiprocessing
import time
import os, sys
from math import sqrt, floor


def isprime(num):
    if num < 2:
        return None
    if num == 2:
        return num
    if num % 2 == 0:
        return None
    for d in range(3, floor(sqrt(num)) + 1, 2):
        if num % d == 0:
            return None
    return num


if __name__ == "__main__":
    MAXNUM = 10 ** int(sys.argv[1])
    THREADS = int(sys.argv[2])
    pool = multiprocessing.Pool(THREADS)
    start_time = time.perf_counter()

    result = list(filter(lambda x: x is not None, pool.map(isprime, range(0, MAXNUM))))

    finish_time = time.perf_counter()
    try:
        # are we using slurm?
        alloc_cpu = os.environ["SLURM_CPUS_PER_TASK"]
    except:
        alloc_cpu = THREADS

    table_data = [
        "threads alloc_cpu primes max_num seconds".split(),
    ]
    table_data.append(
        (THREADS, alloc_cpu, len(result), MAXNUM, round(finish_time - start_time, 4))
    )
    for row in table_data:
        print("{: >20} {: >20} {: >20} {: >20} {: >20}".format(*row))
